/***************************************************************************/
/* Prctica n 3	Gestin de una pila                                      */
/*                                                                         */
/* Fecha: 5 Noviembre 1996                                                 */
/*                                                                         */
/* Autor: Oscar Martnez Valpuesta                                         */
/*                                                                         */
/* Comentarios: Se pretende implementar la rutinas necesarias para la      */
/*	      gestin de una pila: inicializacin, inserccin, extraccin,      */
/*	      eliminacin, etc.                                                 */
/***************************************************************************/
#include <stdio.h> 		/* printf, scanf */
#include <stdlib.h>		/* calloc, exit */
#include <conio.h>		/* clrscr */

/* Macro que devuelve 1 si la pila est vaca. El parmetro que se le pasa
 es el top de la pila.                                                 */
#define PILA_VACIA(X) ((X == NULL) ? 1 : 0)

/* Macro que devuelve el nmero de elementos de la pila. Este n se calcula
 restando al puntero al top el puntero a la base y sumando 1 ya que si solo
 hay un elemento la resta dara 0.                                        */
#define NUMELPILA(X) ((X -> s == NULL) ? 0 : ((unsigned)(X -> s - X -> st) + 1))


typedef int elemento;
typedef struct {
	  elemento *st,		/* puntero a la base */
		   *sb,         /* puntero al lmite superior */
		   *s;          /* puntero al top */
} PILA;

PILA *p;
unsigned n;			/* n de elementos de la pila */
int numero;			/* variable para introducir y extraer */

/* prototipos */
PILA *initPila(unsigned N);
void push(PILA *pila, elemento e);
void pop(PILA *pila, elemento *e);
void Libera_pila(PILA *pila);

/***************************************************************************/
/* Funcin que reserva en primer lugar memoria para la estructura de la pila
 y si esto ha sido posible reserva memoria para los N elementos que contendr
 como mximo.                                                               */
/***************************************************************************/
PILA *initPila(unsigned N) {

  p = (PILA *) malloc (sizeof(PILA));	/* memoria para la estructura */
  if (p == NULL) {
    printf("\nERROR, no se ha podido dar memoria a la pila.");
    exit(1);
  }
  p -> s = (elemento *) calloc (N, sizeof(elemento)); /* memoria para los */
  if (p -> s == NULL) {                               /* N elementos */
    printf("\nERROR, no se ha podido dar memoria a los %d elementos de la pila.", N);
    exit(1);
  }

  /* inicialmente el puntero a la base coincide con el del top */
  p -> st = p -> s;
  p -> sb = (p -> s + N);	/* puntero al lmite */
  p -> s = NULL;
  return p;
}

/****************************************************************************/
/* Funcin que inserta en la pila un elemento e incrementando el top.       */
/****************************************************************************/
void push(PILA *pila, elemento e) {

	(*pila).s++;
	if (pila -> s == pila -> sb) {	/* toda la pila ocupada */
	  printf("\nNo queda ms espacio en la pila.");
	  exit(1);
	}
	*(pila -> s) = e;
}


/***************************************************************************/
/* Funcin que extrae de la pila un elemento e si esta tiene al menos uno. */
/***************************************************************************/
void pop(PILA *pila, elemento *e) {

  if (PILA_VACIA(pila -> s)) {
    printf("\nERROR, la pila est vaca.");
    exit(1);
  }
  *e = *(pila -> s);			/* devuelve el elemento */
  if (NUMELPILA(pila) == 1)
    pila -> s = NULL;
  else
    (*pila).s--;
}

/***************************************************************************/
/* Funcin que libera de la memoria toda la pila, la estructura y lo que se
 haba reservado para los N elementos.                                     */
/***************************************************************************/
void liberapila(PILA *pila) {

  free(pila -> s);
  free(pila);
}

/****************************************************************************/
int main(void) {

  printf("\nIntroduce el nmero de elementos de la pila: ");
  scanf("%d", &n);
  p = initPila(n);

  /* introduccin de nmeros. 0 para parar. */
  do {
    clrscr();
    printf("Nmero a introducir en la pila: ");
    scanf("%d", &numero);
    if (numero != 0)
      push(p, numero);
  } while (numero != 0);

/* extraccin y visualizacin de los elementos hasta que se vace la pila */
  clrscr();
  while (!PILA_VACIA(p -> s)) {
    pop(p, &numero);
    printf("%d ", numero);
  }
  getch();
  liberapila(p);
  return 0;
}