/****************************************************************************
PROGRAMA:	CNJLIB
AUTOR:		Patricio Garca Bez
FECHA:		Curso 92/93
FINALIDAD:	Librera de funciones para trabajar con conjuntos.
COMENTARIOS:	Los conjuntos se implementan como punteros a una estructura
					llamada set (ver cnjlib.h) que contiene

HISTORIA:	Kiko. Adaptada y documentada 28.09.94
				Kiko. Revisin de algn Bug 18.11.94
BIBLIOGRAFIA:
MODO DE UTILIZACION:
****************************************************************************/
#include <alloc.h>
#include "cnjlib.h"


/****************************************************************************
* *cnjini -- 	Aloja memoria para un conjunto con capacidad mxima
*					de num_ele elementos
*
* Parmetros:
*		num_ele: Nmero mximo de elementos que va a almacenarse en el conjunto.
* Devuelve:
*		Un puntero a una estructura de tipo conjunto que se ha inicializado
*		convenientemente.
****************************************************************************/
set *cnjini(unsigned num_ele) {
	palabra *p;
	set *cj;
	unsigned l;

	p = (palabra *) calloc((l = PAL(num_ele - 1) + 1), sizeof(palabra));
	if (p != NULL) {
		cj = (set *) malloc(sizeof(set));
		if (cj == NULL)
			free(p);
		else {
			cj->longitud = l;
			cj->conj = p;
		}
		return (cj);
	} else
		return (NULL);
}
/****************************************************************************
* cnjper -- Indica <>0  0 segun el elemento nmero e pertenezca
*				o no en el conjunto cj1
*
* La pertenencia se da si el bit correspondiente est a 1, lo cual se
* comprueba haciendo un AND de bits con 1 rotado a la izquierda el nmero de
* bits necesario con la palabra correspondiente al elemento e.
*
* Parmetros:
*			*cj -- Puntero al conjunto considerado
*			e   -- Elemento cuya pertenencia se estudia
*****************************************************************************/
int cnjper(set *cj, unsigned e) {
	return((cj->conj[PAL(e)] & (1UL << BIT(e))) != 0);
}
/****************************************************************************
* cnjins -- Aade el elemento nmero e al conjunto cj
*
* Para ello se hace un OR de bits en la palabra y bits adecuados. Si el
* elemento ya perteneca al conjunto, esta operacin no provocar ningn
* cambio.
*
* Parmetros:
*			*cj -- Puntero al conjunto considerado
*			e   -- Elemento a aadir
*****************************************************************************/
void cnjins(set *cj, unsigned e) {
	unsigned i;
	palabra j;

	i = PAL(e);
	j = (1UL << BIT(e));
	cj->conj[i] |= j;
}
/****************************************************************************
* cnjvac -- Inicializa a vaco el conjunto
*
* La operacin se implementa poniendo a 0 todas las palabras del conjunto.
*
* Parmetros:
*			*cj -- Puntero al conjunto considerado
*****************************************************************************/
void cnjvac(set *cj) {
	palabra *i;
	unsigned j;

	j = cj->longitud;
	i = cj->conj;
	while (j--)
		*i++ = 0;
}
/****************************************************************************
* cnjcpy -- Copia el conjunto cjorg en el conjunto cjdest
*
* En la implementacin se est suponiendo implcitamente que el conjunto
* destino tiene una longitud mayor o igual que la del origen.
*
* Parmetros:
*			*cjdest -- Puntero al conjunto destino
*			*cjorg -- Puntero al conjunto origen
*****************************************************************************/
void cnjcpy(set *cjorg, set *cjdest) {
	unsigned k;
	palabra *i, *j;

	k = cjorg->longitud;
	j = cjorg->conj;
	i = cjdest->conj;
	while (k--)
		*i++ = *j++;
	k = (cjdest->longitud) - (cjorg->longitud);
	while (k--)
		*i++ = 0;
}
/****************************************************************************
* cnjigu -- Devuelve <>0  0 segn sean iguales o no los conjuntos cj1 y cj2
*
* En primer lugar se devuelve 0 si el nmero de palabras de los conjuntos es
* distinto. Si son iguales se devuelve 0 en cuanto haya dos palabras distintas
* o bien 1 si todas son iguales.
*
* Parmetros:
*			*cj1 -- Puntero al primer conjunto
*			*cj2 -- Puntero al segundo conjunto
*****************************************************************************/
int cnjigu(set *cj1, set *cj2) {
	unsigned k;
	palabra *i, *j;

	if ((k = cj1->longitud) != (cj2->longitud))
		return(0);
	j = cj1->conj;
	i = cj2->conj;
	while ((--k) && (*i == *j)) {
		i++;
		j++;
	}
	return (*i == *j);
}
/*****************************************************************************
* cnjlib -- Libera la memoria reservada a un conjunto mediante cnjini
*
* Parmetros:
*			*cj -- Puntero al conjunto
*****************************************************************************/
void cnjlib(set *cj) {
	free(cj->conj);
	free(cj);
}
/****************************************************************************
* cnjuni -- Aade los elelmentos del conjunto cjorg en el conjunto cjdest
*
* La unin se implementa como un OR de bits de las palabras de ambos conjuntos
*
* Parmetros:
*			*cjdest -- Puntero al conjunto destino
*			*cjorg -- Puntero al conjunto origen
*****************************************************************************/
void cnjuni( set *cjdest, set *cjorg) {
	unsigned k;
	palabra *i, *j;

	k = cjorg->longitud;
	j = cjorg->conj;
	i = cjdest->conj;
	while (k--)
		*i++ |= *j++;
}
