program conversionDeExpresionRegularAAutomataFinitoNoDeterministico;
{
PROGRAMA:			ER2NFA
AUTOR:				Casiano Rodrguez
FECHA:				Curso 92/93
FINALIDAD:			Generacin de un NFA reconocedor de una expresin regular.
COMENTARIOS:		Acepta expresiones regulares en la entrada y, mediante
						el algoritmo de Thompson obtiene el autmata finito no
						determinista equivalente.
				 		Previamente es necesario analizar sintcticamente las
						expresiones para construir su rbol.
						El autmata se escribe en el fichero pasado como parmetro,
						en la lnea de comandos o bien en 'AUTOMATA.NFA'.
						Para acabar se introduce ^Z.
						El fichero de salida tiene la estructura:
							estado de arranque              
							estado de aceptacion            
							estado numeroDeTransiciones caracter estado ... 		
					             ...                 		
HISTORIA:	    Versin con estados consecutivos.
BIBLIOGRAFIA:   Aho, Sethi y Ullman
					 "Compilers. Principles, techniques and tools"
					 Addison Wesley, 1986
MODO DE UTILIZACION:  ER2NFA [fichero_de_salida]
								Introducir una expresin regular.
}
const
  epsilon = '';                   (* Caracter que representa a epsilon *)
  maxByte = 255;
  LF = #10; (* line feed *)
  CR = #13; (* carriage return *)
  EOF = ^Z;
type
  charSet = set of char;
  ptrNodo = ^nodo;
  nodo = record                    			(* Nodos para almacenar rbol sintctico *)
    case token: char of                   (* .   concatenacion   *)
    '', '|': (left, right: ptrNodo);     (* |   alternativa     *)
    '*': (child: ptrNodo);           		(* *   repeticion      *)
  end (* nodo *);

  ptrEstado = ^estado;

  transicion = record             	(* Estructura para almacenar trancisiones *)
    c: char;                        (* simbolo de entrada  *)
    p: ptrEstado;                   (* estado siguiente    *)
  end;

  estado = record                  		(* Estructura para almacenar estados      *)
    name: byte;                      	(* nombre del estado             *)
    num: byte;                       	(* numero de transiciones        *)
    et: array [0..1] of transicion; 	(* a lo mas hay dos transiciones *)
  end;                                 (* de salida.                    *)

var
  lookahead: char;
  regularAlphabet, alphabet, firstFactor: charset;
  raiz: ptrNodo;
  start, final: ptrEstado;
  nN: byte;
  sal: text;

procedure imprime(s, f: ptrEstado);
(* Imprime en pantalla y fichero de salida el autmata generado *)
var
  procesado: set of byte;

  procedure imp(s: ptrEstado);
  var i: byte;
  begin
    if s <> nil then begin
      with s^ do begin
        write(name:3, ': '); write(sal, name:3, ' ', num, ' ');
        for i := 1 to num do begin
          write(et[pred(i)].c:2, ' = ', et[pred(i)].p^.name:3, ' ');
          write(sal, et[pred(i)].c, et[pred(i)].p^.name:3, ' ');
        end;
        writeln; writeln(sal);
        procesado := procesado+[name];
        for i := 1 to num do
          if not (et[pred(i)].p^.name in procesado) then imp(et[pred(i)].p);
      end (* with *);
    end;
  end (* imp *);
begin (* imprime *)
  if s <> nil then begin
    procesado := [];
    writeln('Estado de arranque: ', s^.name);
    writeln(sal, nN);
    writeln(sal, s^.name);
    writeln('Estado de aceptacion: ', f^.name);
    writeln(sal, f^.name);
    imp(s);
  end;
  {writeln('-------fin de la tabla-------');}
end (* imprime *);

procedure error(s: string);
begin
  writeln('Error!. ', s); halt(1);
end (* error *);

function lex: char;
(* Analizador lxico *)
var car: char;
begin
  read(car);
  if (car = CR) then begin (* en DOS eoln es doble caracter *)
    read(car); (* #10 Cambiar si UNIX. *)
  end
  else
  if not(car in regularAlphabet) then
    error('caracter ilegal');
  lex := car;
end;

function expresionRegular: ptrNodo;
(* Analizador sintctico de expresiones regulares *)
  var p, q: ptrNodo;
  function termino: ptrNodo;
    var p, q: ptrNodo;
    function factor: ptrNodo;
    var p, q: ptrNodo;
    begin
      if (lookahead = '(') then begin
        lookahead := lex;
        p := expresionRegular;
        if (lookahead = ')') then begin
          lookahead := lex;
        end
        else error('parentesis sin casar.');
      end
      else
      if (lookahead = epsilon) then begin
        new(p);
        p^.token := epsilon;
        lookahead := lex;
      end
      else
      if (lookahead in Alphabet) then begin
        new(p); p^.token := lookahead;
        lookahead := lex;
      end
      else error('se esperaba letra o parentesis.');
      if lookahead = '*' then begin
        new(q);
        q^.token := '*';
        q^.child := p;
        p := q;
        lookahead := lex;
        while (lookahead = '*') do begin
          lookahead := lex;
        end;
      end;
      factor := p
    end (* factor *);

  begin (* termino *)
    p := factor;
    if (lookahead in firstFactor) then begin
      new(q);
      q^.token := '';
      q^.left := p;
      q^.right := termino;
      p := q;
    end;
    termino := p
  end (* termino *);

begin (* expresion regular *)
  p := termino;
  if (lookahead = '|') then begin
    new(q);
    q^.token := '|';
    q^.left := p;
    lookahead := lex;
    q^.right := expresionRegular;
    p := q;
  end;
  expresionRegular := p
end (* expresion regular *);

procedure construccionDeThomson(var inicial, final: ptrEstado; r: ptrNodo);
(* Algoritmo de Thomson para construir en NFA *)
  var s, f: ptrEstado;


  procedure nuevoNombre(var n: byte);
  (* Asigna nombres no ocupados a los estados *)
  begin
    if nN < maxByte then n := nN
    else error('Expresion demasiado compleja.');
    inc(nN);
  end (* nuevoNombre *);

  procedure liberaNombre;
  (* Libera el ltimo nombre utilizado *)
  begin
    nN := nN - 1;
  end (* liberaNombre *);

begin
  case r^.token of
    '|': begin              (* Contruccin de alternativas *)
      new(inicial);
      new(final);


      final^.num := 0;

      inicial^.num := 2;

      construccionDeThomson(s, f, r^.left);

      inicial^.et[0].c := epsilon;
      inicial^.et[0].p := s;

      f^.num := 1;
      f^.et[0].p := final;
      f^.et[0].c := epsilon;

      construccionDeThomson(s, f, r^.right);

      inicial^.et[1].c := epsilon;
      inicial^.et[1].p := s;

      f^.num := 1;
      f^.et[0].p := final;
      f^.et[0].c := epsilon;

      nuevoNombre(final^.name);
      nuevoNombre(inicial^.name);

    end;
    '': begin              (* Construccin de concatenaciones *)
      construccionDeThomson(s, final, r^.right);
      liberaNombre;
      construccionDeThomson(inicial, f, r^.left);
      f^.num := s^.num; f^.et := s^.et;
      dispose(s);
    end;                     (* Construccin de repeticiones *)
    '*': begin
      new(inicial);
      new(final);
      final^.num := 0;
      inicial^.num := 2;
      inicial^.et[1].c := epsilon;
      inicial^.et[1].p := final;
      construccionDeThomson(s, f, r^.child);
      inicial^.et[0].c := epsilon;
      inicial^.et[0].p := s;
      f^.num := 2;
      f^.et[0].c := epsilon;
      f^.et[0].p := s;
      f^.et[1].c := epsilon;
      f^.et[1].p := final;

      nuevoNombre(final^.name);
      nuevoNombre(inicial^.name);
    end;
    else begin             (* Construccin de epsilon o symbol *)
      new(inicial); new(final);
      nuevoNombre(final^.name);
      nuevoNombre(inicial^.name);
      final^.num := 0;
      inicial^.num := 1;
      inicial^.et[0].c := r^.token;
      inicial^.et[0].p := final;
    end;
  end (* case *);
end;

 begin (* main *)
  writeln('Introduzca Expresiones regulares seguidas de RETURN');
  writeln('Pulse CTRL Z para finalizar la introduccin de expresiones.');
  writeln('El fichero de descripcin del NFA generado se encuentra en el fichero');
  writeln('que se ha pasado como parmetro en lnea de comandos, o bien en el fichero');
  writeln('AUTOMATA.NFA en caso que no se haya pasado parmetro.');

  regularAlphabet := [' '..epsilon, CR, ^Z];
  alphabet := regularAlphabet-['|',  '(', ')', '*', epsilon];
  firstFactor := alphabet+[epsilon, '('];
  if (ParamCount>0) then
    assign(sal, ParamStr(1))
  else
    assign(sal, 'AUTOMATA.NFA');
  rewrite(sal);
  lookahead := lex;
  repeat
    raiz := expresionRegular;
    if (lookahead = LF) then begin
      writeln('ok. ');
      nN :=  0;
      construccionDeThomson(start, final, raiz);
      imprime(start, final);
      lookahead := lex;
    end
    else error('expresion regular mal terminada.');
  until (lookahead = EOF);
  close(sal);
end.

