/****************************************************************************
 Prctica n 10	   Construccin del rbol sintctico de una expresin
		   regular.

 Fecha: 14 Enero 1997

 Autor: Oscar Martnez Valpuesta

 Objetivo:   Se pretende aplicar el mtodo de anlisis sintctico descendente
	   recursivo para crear el rbol sintctico asociado a una expresin
	   regular correcta.
 Comentarios: La gramtica de las expresiones regulares que se utilizar es:

	    Expresin --->Trmino Resto_expresin
	    Resto_expresin ---> | Expresin
	    Resto_expresin ---> 
	    Trmino ---> Cotrmino Resto_trmino
	    Resto_trmino ---> Trmino
	    Resto_trmino --> 
	    Cotrmino ---> Factor Resto_cotrmino
	    Resto_cotrmino ---> * Resto_cotrmino
	    Resto_cotrmino ---> 
	    Factor ---> (Expresin)
	    Factor ---> id
	    Factor ---> &

	    El smbolo & indica la expresin regular vaca y los identificadores
	    son de una letra.

	      Se ha diseado una funcin para cada smbolo no terminal de
	    la gramtica que funcionar como una regla de produccin. Si en
	    el anlisis de una determinada expresin se produce un error se
	    para el proceso y se visualiza la posicin en la cadena del error.

	      Las cadenas de entrada se leen desde un fichero de texto pasado
	    por parmetro al programa.

	      Para obtener el rbol sintctico se utilizaran las siguientes
	    reglas semnticas asociadas a las reglas sintcticas arriba
	    citadas:

	    Expresin.A --> nodo (.)(Trmino.A, Resto_Expresin.A )
	    Resto_expresin.A --> nodo (|)(NULL, Expresion.A)
	    Resto_expresin.A --> NULL
	    Trmino.A --> nodo (.)(Cotermino.A, Resto_trmino.A)
	    Resto_trmino.A --> Trmino.A
	    Resto_trmino.A --> NULL
	    Cotrmino.A --> nodo (.)(Factor.A, Resto_cotrmino.A)
	    Resto_cotrmino.A --> nodo (*)(NULL, Resto_cotrmino.B)
	    Resto_cotrmino.A --> NULL
	    Factor.A --> nodo (Expresin.A)(NULL, NULL)
	    Factor.A ---> nodo (Id)(NULL, NULL)
	    Factor.A --> nodo (&)(NULL, NULL)
/****************************************************************************/
#include <string.h>    /* strlen */
#include <stdlib.h>
#include "util.h"

#define MAX 40	   /* n de elementos como mximo en la cadena de entrada */

/******************************* TIPOS *************************************/
/* conjunto de tokens posibles en el anlisis */
/* IDENT es un identificador (una letra en minscula) */
/*                    *     (       )   |    &           $            */
typedef enum {IDENT, AST, A_PAR, C_PAR, OR, AND, ERROR, FIN} tokens;
typedef struct tipo_nodo {     /* estructura de los nodos del rbol */
  int dato;
  struct tipo_nodo *izq;       /* tiene mayor prioridad que su hermano */
  struct tipo_nodo *dcha;
} nodo;

/****************************** VARIABLES **********************************/
FILE *fichero;		/* handled para el fichero de entrada */
char cadena[MAX];       /* cadena de entrada */
unsigned posicion;	/* posicin actual en la cadena de entrada */
unsigned longitud;	/* longitud de la cadena */
tokens lookahead;	/* smbolo de preanlisis */
nodo *arbol;            /* puntero a la raiz del rbol */


/***************************** PROTOTIPOS **********************************/
void Abrir(FILE **f, char s[]);
void Leer_cadena(char *s);
tokens get_token(unsigned p);
void match(tokens token);
nodo *Expresion(void);
nodo *Termino(void);
nodo *Resto_expresion(void);
nodo *Cotermino(void);
nodo *Resto_termino(void);
nodo *Factor(void);
nodo *Resto_cotermino(nodo *nodo2);
void Visualizar (nodo *arbol, int cont);
void Liberar (nodo *arbol);

/***************************************************************************/
/* Funcin que abre un fichero de alto nivel para lectura.                 */
/***************************************************************************/
void Abrir(FILE **f, char s[]) {

  *f = fopen(s, "r");
  if (*f == NULL) {
    ERROR1(s);
    exit(0);
  }
}

/***************************************************************************/
/*  Funcin que lee desde un fichero de texto una cadena de entrada y      */
/* almacena en una variable global la longitud de la misma.                */
/***************************************************************************/
void Leer_cadena(char *s) {

  fscanf(fichero, "%s", s);
  longitud = strlen(s);
  posicion = 0;
}

/***************************************************************************/
/* Funcin que devuelve el token de la posicin p en la cadena de entrada. */
/***************************************************************************/
tokens get_token(unsigned p) {

  if ((cadena[p] >= 'a') && (cadena[p] <= 'z'))
    return(IDENT);

  switch (cadena[p]) {
    case '*': return(AST);
    case '|': return(OR);
    case '&': return(AND);
    case '(': return(A_PAR);
    case ')': return(C_PAR);
    case '\0': return(FIN);
    default: return(ERROR);
  }
}

/***************************************************************************/
/*  Funcin que actualiza el smbolo de preanlisis siempre y cuando no se */
/* haya producido un error y coincida el carcter esperado con el lookahead*/
/* actual.         							   */
/***************************************************************************/
void match(tokens token) {

  if (lookahead == token)
    lookahead = get_token(++posicion);
  else
    lookahead= ERROR;
}

/***************************************************************************
Expresin --> Trmino Resto_expresin
****************************************************************************/
nodo *Expresion(void) {
nodo *nodo1, *nodo2;

  nodo1 = Termino();
  nodo2 = Resto_expresion();
  if (nodo2 != NULL) {
    nodo2 -> izq = nodo1;
    return (nodo2);
  }
  return (nodo1);
}

/***************************************************************************
 Trmino --> Cotrmino Resto_trmino
****************************************************************************/
nodo *Termino(void) {
nodo *nodo1, *nodo2;

  nodo1 = Cotermino();
  nodo2 = Resto_termino();
  if (nodo2 != NULL) {
    nodo2 -> izq = nodo1;
    return(nodo2);
  }
  return(nodo1);
}

/***************************************************************************
 Resto_expresin --> | Expresin
 Resto_expresin --> 
****************************************************************************/
nodo *Resto_expresion(void) {
nodo *nodo1;

  nodo1 = NULL;  /* pude ser la palabra vaca */
  if (lookahead == OR) {
    match(OR);
    nodo1 = (nodo *) malloc(sizeof(nodo));
    if (nodo1 == NULL) {
      ERROR_MEM;
      lookahead = ERROR;
      return(NULL);
    }
    nodo1 -> dato = '|';
    nodo1 -> izq = NULL;
    nodo1 -> dcha = Expresion();
  }
  return(nodo1);
}

/***************************************************************************
  Cotrmino --> Factor Resto_cotrmino
****************************************************************************/
nodo *Cotermino(void) {
nodo *nodo1, *nodo2;

  nodo1 = Factor();
  if (lookahead != ERROR) {
    nodo2 = Resto_cotermino(nodo1);
    return(nodo2);
  }
  return(nodo1);
}

/***************************************************************************
 Resto_trmino --> Trmino
 Resto_trmino --> 
****************************************************************************/
nodo *Resto_termino(void) {
nodo *nodo1;

  nodo1 = NULL;   /* palabra vaca */
  /* verifica que el token pertenece al first de Trmino */
  if ((lookahead == A_PAR) || (lookahead == AND) || (lookahead == IDENT)) {
    nodo1 = (nodo *) malloc(sizeof(nodo));
    if (nodo1 == NULL) {
      ERROR_MEM;
      lookahead = ERROR;
      return(NULL);
    }
    nodo1 -> dato = '.';
    nodo1 -> izq = NULL;
    nodo1 -> dcha = Termino();
  }
  return (nodo1);
}

/***************************************************************************
 Factor --> (Expresin)
 Factor --> id
 Factor --> &
****************************************************************************/
nodo *Factor(void) {
nodo *nodo1;

  nodo1 = (nodo *) malloc(sizeof(nodo));
  if (nodo1 == NULL) {
    ERROR_MEM;
    lookahead = ERROR;
    return(NULL);
  }
  nodo1 -> izq = NULL;
  nodo1 -> dcha = NULL;
  if (lookahead == A_PAR) {
    match(A_PAR);
    free(nodo1);
    nodo1 = Expresion();
    match(C_PAR);
  }
  else
    if (lookahead == AND) {
      nodo1 -> dato = '&';
      match(AND);
    }
    else
      if (lookahead == IDENT) {
	nodo1 -> dato = cadena[posicion];
	match(lookahead);
      }
      else
	lookahead = ERROR;
  return(nodo1);
}

/***************************************************************************
 Resto_cotrmino --> * Resto_cotrmino
 Resto_cotrmino --> 
****************************************************************************/
nodo *Resto_cotermino(nodo *nodo2) {
nodo *nodo1;

  nodo1 = nodo2;
  if (lookahead == AST) {
    match(AST);
    nodo1 = (nodo *) malloc(sizeof(nodo));
    if (nodo1 == NULL) {
      ERROR_MEM;
      lookahead = ERROR;
      return(NULL);
    }
    nodo1 -> dato = '*';
    nodo1 -> dcha = NULL;
    nodo1 -> izq = Resto_cotermino(nodo2);
  }
  return (nodo1);
}

/***************************************************************************
   Este procedimiento se encarga de visualizar un determinado rbol
  sintctico asociado a una expresin regular. Recibe como primer parmetro
  el puntero al nodo raiz del rbol, y como segundo, un contador encargado
  de controlar los espacios en blanco con el fin de visualizar el rbol de
  forma que se pueda entender.
****************************************************************************/
void Visualizar(nodo *arbol, int i) {
int blancos;

  if (arbol != NULL) {
    Visualizar(arbol -> dcha, (i + 2));
    blancos = i;
    printf ("\n");
    while (blancos > 0) {
      printf ("  ");
      blancos--;
    }
    printf("%c", arbol -> dato);
    Visualizar(arbol -> izq, (i + 2));
  }
}

/****************************************************************************
   Este procedimiento se encarga de liberar la memoria que ha sido reservada
 en el proceso de construccin de un determinado rbol sintctico.
****************************************************************************/
void Liberar(nodo *arbol) {

  if (arbol != NULL) {
    Liberar(arbol -> izq);
    Liberar(arbol -> dcha);
    free(arbol);
  }
}

/***************************************************************************/
int main(int argc, char *argv[]) {

  clrscr();
  if (argc != 2){
    printf("\nERROR, forma de uso: ejecutable <archivo_de_entrada.IN>\n");
    exit(0);
  }
  Abrir(&fichero, argv[1]);	/* apertura del fichero de datos */
  while (!feof(fichero)) {
    Leer_cadena(cadena);	/* leemos una cadena de entrada */
    lookahead = get_token(posicion);  /* inicializacin del lookahead */
    arbol = Expresion();	      /* se inicia el anlisis */
    if (lookahead == FIN) {
      printf("\n\n%s\tExpresin regular correcta.\n", cadena);
      Visualizar(arbol, 1);
      Pausa();
    }
    else {
      printf("\n\n%s\tExpresin regular incorrecta, error en la posicin %d.", cadena, posicion+1);
      Pausa();
    }
    Liberar(arbol);
  }

  fclose(fichero);
  return 0;
}
