/***************************************************************************/
/* Prctica n 2	ORDENACION DE FICHEROS                             */
/*                                                                         */
/* Fecha: 29 Octubre 1996                                                  */
/*                                                                         */
/* Autor: Oscar Martnez Valpuesta                                         */
/*                                                                         */
/* Propsito:   Este programa genera un fichero de registros que ms tarde */
/*	      ser ordenado por el mtodo de mezcla natural.               */
/*                                                                         */
/* Comentarios: El programa genera el campo clave aleatoriamente y ordena  */
/*	      mediante el Quicksort tantos elementos como le sea posible   */
/*	      para almacenarlos luego en un fichero. El usuario especifca */
/*	      el nmero de registros a crear. Una vez generado el fichero  */
/*	      de entrada, la funcin Mezcla_natural lo ordena con ayuda de */
/*	      dos ficheros auxiliares.                                     */
/*                                                                         */
/* Definicin de tramo: es una secuencia de nmeros ordenados.             */
/***************************************************************************/

#include <stdio.h>
#include <io.h>		/* open */
#include <fcntl.h>	/* O_BINARY, O_RDWR */
#include <stdlib.h>	/* exit, malloc, calloc */
#include <sys/stat.h>   /* S_IWRITE, S_IREAD */
#include <time.h>	/* clock */

#define Q 230   	/* n mximo de elementos que soporta el Quicksort */

typedef struct {
  unsigned clave;
  char c;
  int i;
  float r;
  char cadena[50];
  int v[100];
} registro;

typedef enum {FALSE = 0, TRUE = 1} boolean;

int max_items;		     /* n mximo de items */
registro *buffer;	     /* ordenar los datos */
registro *buffer2, *buffer3; /* auxiliares para mezclar los datos */
clock_t inicial, final;	     /* instantes inicial y final del proceso */
int n;			     /* n de elementos */
int fichero;		     /* fichero de datos */
int fichero_salida;	     /* fichero de salida despus de mezcla natural */
int num;		     /* n de registros leidos */
int aux;
int iteraciones;
int a,b;		     /* ficheros auxiliares para la mezcla */
boolean fdt = FALSE;	     /* indica fin de tramo */
int l;			     /* n de tramos a mezclar */


/* prototipos */
void Crear(char cadena[], int *f);
void Dar_memoria(void);
void Rellenar(void);
void Rellena(int m);
int Leer(int *f);
void Ordenar (int izq, int der);
void Copia(int x, int y);
void Copia_tramo(int x, int y);
void Mezcla_tramo(void);
void Distribucion(void);
void Mezclar(void);
void Mezcla_natural(void);


/***************************************************************************/
/* Crea un nuevo fichero o lo sobreescribe si este ya exista.             */
/***************************************************************************/
void Crear(char cadena[], int *f) {

  *f = creat(cadena, S_IREAD | S_IWRITE);
  if (*f == -1) {
    printf("\nERROR, fichero %s no abierto correctamente.\n", cadena);
    exit(1);
  }
}

/****************************************************************************/
/* Inicializa el buffer y calcula el n mximo de registros que caben el l */
/****************************************************************************/
void Dar_memoria() {

  max_items = Q;
  buffer = (registro *) calloc(max_items, sizeof(registro));
  while (buffer == NULL) {
    free(buffer);
    max_items--;
    buffer = (registro *) calloc(max_items, (int) sizeof(registro));
  }


}

/***************************************************************************/
/* Funcin se encarga de rellenar tantos registros como sea posible.       */
/***************************************************************************/
void Rellenar() {
register j;

  aux = n;
  for (j = 0; j <= iteraciones; j++) {
    aux = aux - (iteraciones * max_items);
    if ( (aux < max_items) && (aux > 0))
      Rellena(aux);
    else
      Rellena(max_items);
  }
}

/***************************************************************************/
/* Funcin que rellena aleatoriamente m registros. 			   */
/***************************************************************************/
void Rellena(int m) {
register int k;

  for (k = 0; k < m; k++) {
    (* (buffer + k)).clave = rand();
    (* (buffer + k)).c = 'a';
    (* (buffer + k)).i = -1;
    (* (buffer + k)).r = 1.5;
    (* (buffer + k)).cadena[1] = 'b';
    (* (buffer + k)).v[1] = 1;
  }
  write(fichero, buffer, m * sizeof(registro));  /* escritura de m registros */
}

/***************************************************************************/
/* Fucin que lee tantos registros como sea posible y que devuelve el n   */
/* de los que ha leido. Nunca lee ms de el nmero mximo.                 */
/***************************************************************************/

int Leer(int *f) {
int m;

  if (aux > max_items)
    m = max_items;
  else
    m = aux;
  aux = aux - m;
  read(*f, buffer, m * sizeof(registro));
 return m;
}

/***************************************************************************/
/* Funcin que ordena mediante el mtodo QuickSort un vector dado          */
/***************************************************************************/
void Ordenar (int izq, int der) {
register int i, j; 		/* variables ndice del vector */
int medio; 			/* contiene el elemento medio del vector */
registro reg; 			/* variable auxiliar */

  i = izq;
  j = der;
  medio = buffer[(izq + der) / 2].clave;       /* elemento "mediana" */
  do {
    while (buffer[i].clave < medio)
      i++;
    while (medio < buffer[j].clave)
      j--;
    if (i <= j) { 				/* intercambiar */
      reg = buffer[i];
      buffer[i] = buffer[j];
      buffer[j] = reg;
      i++;
      j--;
    }
  } while (i <= j);
  if (izq < j)
    Ordenar (izq, j);                          /* ordenar la parte izquierda */
  if (i < der)
    Ordenar (i, der);                          /* ordenar la parte derecha */
}

/***************************************************************************/
/* Procedimiento que copia un elemento de x en y y calcula si se ha llegado*/
/* al final de un tramo.                                                   */
/***************************************************************************/
void Copia(int x, int y) {

  read(x, buffer2, sizeof(registro));
  write(y, buffer2, sizeof(registro));
  if (eof(x))			      /* el final de fichero indica */
    fdt = TRUE;                       /* el final de un tramo */
  else {
    read(x, buffer3, sizeof(registro));               /* si el elemento */
    lseek(x, -1 * (long)sizeof(registro), SEEK_CUR);  /* siguiente es menor */
    fdt = buffer2 -> clave > buffer3 -> clave;        /* tambin es fdt */
  }
}

/***************************************************************************/
/* Funcin que copia elemtos de x a y hasta llegar al final de un tramo.   */
/***************************************************************************/
void Copia_tramo(int x, int y) {		/* x origen */
						/* y destino */
  do {
    Copia(x, y);
  } while (!fdt);
}

/***************************************************************************/
/* Procedimiento que mezcla dos tramos mientras sea posible. Esta mezcla   */
/* consiste en ir comparando los elementos de los tramos e ir escribindolos*/
/* en otro fichero ya ordenados. Si un tramo se acaba, se copia el resto   */
/* del otro al final.                                                      */
/***************************************************************************/
void Mezcla_tramo() {

  do {
    read(a, buffer2, sizeof(registro));
    read(b, buffer3, sizeof(registro));
    lseek(a, -1 * (long)sizeof(registro), SEEK_CUR);
    lseek(b, -1 * (long)sizeof(registro), SEEK_CUR);
    if (buffer2 -> clave <= buffer3 -> clave) {
      Copia(a, fichero_salida);
      if (fdt)
	Copia_tramo(b, fichero_salida);
    }
    else {
      Copia(b, fichero_salida);
      if (fdt)
	Copia_tramo(a, fichero_salida);
    }
  } while (!fdt);
}

/***************************************************************************/
/* Procedimiento que distribuye por tramos los datos de entrada en dos     */
/* ficheros auxiliares.                                                    */
/***************************************************************************/
void Distribucion(void) {

  do {
    Copia_tramo(fichero_salida, a);
    if (!eof(fichero_salida))
      Copia_tramo(fichero_salida, b);
  } while (!eof(fichero_salida));
}

/***************************************************************************/
/* Mezcla los elementos de los ficheros auxiliares. Si uno de los dos      */
/* ficheros finaliza el resto del otro se copia al final del de salida     */
/***************************************************************************/
void Mezclar(void) {

  while (!eof(a) && !eof(b)) {
    Mezcla_tramo();
    l++;
  }
  while (!eof(a)) {
    Copia_tramo(a, fichero_salida);
    l++;
  }
  while (!eof(b)) {
    Copia_tramo(b, fichero_salida);
    l++;
  }
}

/***************************************************************************/
/* Procemiento que ordena los datos en un fichero. Finaliza cuando solo    */
/* quede un tramo, que estar ordenado. Se basa principalmente en dos      */
/* fases: una de distribucin y otra de mezcla.                            */
/***************************************************************************/
void Mezcla_natural(void) {


  buffer2 = (registro *) malloc(sizeof(registro));
  buffer3 = (registro *) malloc(sizeof(registro));
  if (buffer2 == NULL || buffer3 == NULL) {
    printf("\nERROR, no se ha podido dar memoria al buffer .\n");
    exit(1);
  }
  do {
    /* creamos los ficheros auxiliares */
    Crear("c:\\a.dat", &a);
    Crear("c:\\b.dat", &b);
    close(fichero_salida);
    fichero_salida = open("c:\\salida.dat", O_BINARY | O_RDWR);

    Distribucion();

    /* posicionarnos al principio en los auxiliares y crear el destino */
    close(a);
    close(b);
    a = open("c:\\a.dat", O_BINARY | O_RDWR);
    b = open("c:\\b.dat", O_BINARY | O_RDWR);
    close(fichero_salida);
    Crear("c:\\salida.dat", &fichero_salida);

    l = 0;
    Mezclar();
  } while (l != 1);
  free(buffer2);
  free(buffer3);
  close(a);
  close(b);
}

/***************************************************************************/
int main (void) {
register h;

  _fmode = O_BINARY;

  /* fichero que contiene los registro desordenados */
  Crear("c:\\entrada.dat", &fichero);

  printf("\nIntroduce el nmero de elementos: ");
  scanf("%d", &n);
  Dar_memoria();

  /* clculo del n de iteraciones a realizar para generar un fichero de n */
  /* elementos teniendo en cuenta que en cada pasada solo se pueden generar */
  /* max_items */
  iteraciones = ((n % max_items) == 0) ? (n / max_items) - 1 : (n / max_items);

  Rellenar();
  close(fichero);
  fichero = open("c:\\entrada.dat", O_BINARY | O_RDONLY);

  /* fichero que contiene todos los registros ordenados por el campo clave */
  Crear("c:\\salida.dat", &fichero_salida);

  /* vamos a leer en varias pasadas los n elementos, a ordenarlos y a */
  /* escribirlos en el fichero de salida */
  aux = n;
  for (h = 0; h <= iteraciones; h++) {
    num = Leer(&fichero);
    Ordenar(0, num - 1);
    write(fichero_salida, buffer, num * sizeof(registro));
  }

  free(buffer);
  close(fichero);

  /* vamos a calcular el tiempo que tarda la mezcla natural */
  inicial = clock();
  Mezcla_natural();
  final = clock();

  printf("\nTiempo empleado: %f", (final - inicial) / CLK_TCK);
  close(fichero_salida);
  return(0);
}
