/****************************************************************************/
/* Prctica n 4	Simulacin de un DFA                                */
/*                                                                          */
/* Fecha: 12 Noviembre 1996                                                 */
/*                                                                          */
/* Autor: Oscar Martnez Valpuesta                                          */
/*                                                                          */
/* Comentarios: Se pretende simular el funcionamiento de autmata finito    */
/*		determinista. El programa toma desde la lnea de comandos   */
/*		dos ficheros: el primero contiene la informacin del autmata*/
/*		y el segundo las cadenas de entrada.                        */
/*		  Gracias a la forma de trabajar C con punteros nos va a    */
/*		permitir usar una estructura de datos muy simple para simular*/
/*		el DFA. En primer lugar tenemos un puntero a punteros (**dfa)*/
/*		que lo podemos ver como un array en el que cada celda va a  */
/*		ser un puntero a un estado del autmata. Dependiendo de la  */
/*		entrada este transitar a un estado u otro. Esta decisin se */
/* 		contempla en el campo tran, que es otro puntero a la lista  */
/*		de transiciones de ese estado.                              */
/****************************************************************************/

#include <stdio.h>		/* printf, fscanf, fopen, feof */
#include <conio.h>		/* getch */
#include <stdlib.h>		/* exit, calloc */
#include <string.h>		/* strlen */

#define MAXSTR 20 		/* tamao de las frases de entrada */

typedef enum {FALSE = 0, TRUE = 1} boolean;

typedef struct {
  char entrada;			/* entrada con que se transita */
  unsigned nuevo_estado;	/* nuevo estado al que se pasa */
} transicion;

typedef struct {
  unsigned num_transis;
  transicion *tran;
  boolean aceptacion;
} estados;

FILE *fich;
estados **dfa;					/* estructura del DFA */
unsigned num_estados;		/* n de estados del DFA */
unsigned e_inicial;			/* estado inicial del DFA */

/* prototipos */
void Abrir(FILE **f, char s[]);
void Visualizar(void);
void Leer_DFA(void);
void Simulacion(void);

/***************************************************************************/
/* Funcin que abre un fichero de alto nivel para lectura.                 */
/***************************************************************************/
void Abrir(FILE **f, char s[]) {

  *f = fopen(s, "r");
  if (*f == NULL) {
    printf("\nERROR, el fichero %s no se ha podido abrir.", s);
    exit(0);
  }
}

/***************************************************************************/
/* Procedimiento que muestra en pantalla las caractersticas del autmata. */
/***************************************************************************/
void Visualizar(void) {
register unsigned i, j;

  for (i = 0; i < num_estados; i++) {
	 printf("\nEstado: %d\nAceptacin: %d\nN de transiciones: %d", i, (*dfa[i]).aceptacion, (*dfa[i]).num_transis);
	 for (j = 0; j < (*dfa[i]).num_transis; j++)
		printf("\n\tEntrada: %c, Nuevo estado: %d", (*dfa[i]).tran[j].entrada, (*dfa[i]).tran[j].nuevo_estado);
  }
  printf("Pulse una tecla para continuar...");
  getch();
}
/***************************************************************************/
/* Fucin encargada de crear la estructura de datos del DFA.               */
/***************************************************************************/
void Leer_DFA(void) {
register unsigned i, j;
unsigned k;									/* estado actual */

  fscanf(fich, "%d", &num_estados);	/* lectura del n total de estados */
  fscanf(fich, "%d", &e_inicial);	/* lectura del estado inicial */

  /* memoria para la estructura */
  dfa = (estados **) malloc(num_estados * sizeof(estados));
  if (dfa == NULL) {
	 printf("\nERROR, no se ha podido dar memoria a la estructura.");
	 exit(0);
  }
  for (j = 0; j < num_estados; j++) {
	 fscanf(fich, "%d", &k);			/* lectura estado actual */

	 /* memoria para el estado actual */
	 dfa[k] = (estados *) malloc(sizeof(estados));
	 if (dfa[k] == NULL) {
		printf("\nERROR, no se ha podido dar memoria al registro.");
		exit(0);
	 }
	 fscanf(fich, "%d", &dfa[k] -> aceptacion);
	 fscanf(fich, "%d", &dfa[k] -> num_transis);

	 /* memoria para los registros de transiciones */
	 if (dfa[k] -> num_transis != 0) {
		dfa[k] -> tran = (transicion *) calloc(dfa[k] -> num_transis, sizeof(transicion));
		if (dfa[k] -> tran == NULL) {
		  printf("\nERROR, no se ha podido dar memoria a las transiciones.");
		  exit(0);
		}

		/* lectura de las transiciones del estado actual */
		for (i = 0; i < dfa[k] -> num_transis; i++) {
		  fscanf(fich, " %c", &dfa[k] -> tran[i].entrada);
		  fscanf(fich, "%d", &dfa[k] -> tran[i].nuevo_estado);
		}
	 }
  }
}

/***************************************************************************/
/* Funcin que simula el funcionamiento del DFA. Las cadenas de entrada se */
/* obtienen de un fichero. El procemiento muestra en cada instante el estado*/
/* en que se encuentra, la entrada actual y el estado al que transita.     */
/* Tambin contempla si el estado al que se transita es un estado de muerte*/
/* con lo que el autmata se detiene y pasa a ota cadena de entrada. Al    */
/* final se visualiza si el estado en que queda el autmata es de acptacin*/
/* o no.                                                                   */
/***************************************************************************/
void Simulacion(void) {
register unsigned i, j;
char string[MAXSTR];			/* cadena de entrada */
char ch;							/* caracter de entrada */
unsigned estado;				/* estado actual */
boolean acepta;				/* TRUE si la entrada es aceptada por el estado */
boolean estado_muerte;		/* TRUE si se ha alcanzado este estado */


  while (!feof(fich)) {
	 fscanf(fich, "%s", string);		/* lectura de la cadena */
	 estado_muerte = FALSE;
	 estado = e_inicial;
	 printf("\nCadena: %s\n", string);
	 printf("\nEstado actual\tEntrada\tNuevo estado\n");

	 /* computacin de cada entrada */
	 for (i = 0; i < strlen(string) && !estado_muerte; i++) {
		ch = string[i];
		acepta = FALSE;

		/* ver si el estdo actual acepta la entrada */
		for (j = 0; j < dfa[estado] -> num_transis && !acepta; j++)
		  if (dfa[estado] -> tran[j].entrada == ch)
			 acepta = TRUE;
		j--;

		if (acepta) {		/* caracter de entrada vlido */
		  printf("\t%d\t   %c\t     %d\n", estado, ch, dfa[estado] -> tran[j].nuevo_estado);
		  estado = dfa[estado] -> tran[j].nuevo_estado;  /* cambio de estado */
		}
		else {	/* caracter de entrada no vlido => estado de muerte */
		  printf("\t%d\t   %c\testado de muerte\n", estado, ch);
		  estado_muerte = TRUE;
		}
	 }

	 if (!estado_muerte && dfa[estado] -> aceptacion)
		printf("\nEstado actual: ACEPTACION\n\n");
	 else
		printf("\nEstado actual: NO ACEPTACION\n\n");

	 printf("Pulse una tecla para continuar...");
	 getch();
  }
}

/***************************************************************************/
int main(int argc, char *argv[]) {

  clrscr();
  if (argc != 3){
	 printf("\nERROR, forma de uso: archivo_de_definicin.DFA archivo_de_entrada.IN\n");
	 exit(0);
  }
  Abrir(&fich, argv[1]);	      /* apertura fichero de definicin */
  Leer_DFA();			      /* caractersticas del autmata */
  Visualizar();
  Abrir(&fich, argv[2]);
  Simulacion();     		      /* simulacin del funcionamiento */
  return 0;
}

