/****************************************************************************/
/* Prctica n 5	Simulacin de un NFA                                */
/*                                                                          */
/* Fecha: 19 Noviembre 1996                                                 */
/*                                                                          */
/* Autor: Oscar Martnez Valpuesta                                          */
/*                                                                          */
/* Comentarios: Se pretende simular el funcionamiento de autmata finito no */
/*		determinista. El programa toma desde la lnea de comandos   */
/*		dos ficheros: el primero contiene la informacin del autmata*/
/*		y el segundo las cadenas de entrada.                        */
/*		  Gracias a la forma de trabajar C con punteros nos va a    */
/*		permitir usar una estructura de datos muy simple para simular*/
/*		el NFA. En primer lugar tenemos un puntero a punteros (**nfa)*/
/*		que lo podemos ver como un array en el que cada celda va a  */
/*		ser un puntero a un estado del autmata. Dependiendo de la  */
/*		entrada este transitar a un estado u otro. Esta decisin se*/
/* 		contempla en el campo tran, que es otro puntero a la lista  */
/*		de transiciones de ese estado.                              */
/****************************************************************************/

#include "pila.c"		/* incluye stdio, stdlib, conio */
#include "cnjlib.c"		/* funciones de conjuntos */
#include <string.h>		/* strlen */

#define MAXSTR 20 		/* tamao de las frases de entrada */

typedef enum {FALSE = 0, TRUE = 1} boolean;

typedef struct {
  char entrada;			/* entrada con que se transita */
  unsigned nuevo_estado;	/* nuevo estado al que se pasa */
} transicion;

typedef struct {
  unsigned num_transis;          /* n de transiciones del estado */
  transicion *tran;		 /* lista de transiciones */
} estados;

FILE *fich;			/* handler para los ficheros */
estados **nfa;			/* estructura del NFA */
unsigned num_estados;		/* n de estados del NFA */
unsigned e_inicial;		/* estado de arranque del NFA */
unsigned e_aceptacion;		/* estado de aceptacin */

/* prototipos */
void Pausa(void);
void Abrir(char s[]);
void Visualizar(void);
void Mostrar(set conjunto);
void Leer_NFA(void);
void Clausura(PILA *p, set *conjunto);
void move(set *conjunto, char a);
void Simulacion(void);

/***************************************************************************/
void Pausa(void) {

  printf("\nPulse una tecla para continuar...");
  getch();
}

/***************************************************************************/
/* Funcin que abre un fichero de alto nivel para lectura.                 */
/***************************************************************************/
void Abrir(char s[]) {

  fich = fopen(s, "r");
  if (fich == NULL) {
    printf("\nERROR, el fichero %s no se ha podido abrir.", s);
    exit(0);
  }
}

/***************************************************************************/
/* Funcin que muestra en pantalla el contenido de un conjunto pasado por  */
/* parmetro.                                                              */
/***************************************************************************/
void Mostrar(set conjunto) {
register unsigned i;

  printf("{");
  for (i = 0; i < num_estados; i++)
    if (cnjper(&conjunto, i))
      printf("%d ", i);
  printf("}");
}

/***************************************************************************/
/* Procedimiento que muestra en pantalla las caractersticas del autmata: */
/* estado, nmero de transiciones, lista de transiciones segn la entrada. */
/***************************************************************************/
void Visualizar(void) {
register unsigned i, j;

  for (i = 0; i < num_estados; i++) {
    printf("\n[%d]  %d  ", i, nfa[i] -> num_transis);
    for (j = 0; j < nfa[i] -> num_transis; j++)
      printf("%c %d ", nfa[i] -> tran[j].entrada, nfa[i] -> tran[j].nuevo_estado);
  }
 Pausa();
}
/***************************************************************************/
/* Fucin encargada de crear la estructura de datos del NFA.               */
/***************************************************************************/
void Leer_NFA(void) {
register unsigned i, j;
unsigned k;				/* estado actual */

  fscanf(fich, "%d", &num_estados);	/* lectura del n total de estados */
  fscanf(fich, "%d", &e_inicial);	/* lectura del estado inicial */
  fscanf(fich, "%d", &e_aceptacion);	/* lectura del estado de aceptacin */

  /* memoria para la estructura */
  nfa = (estados **) malloc(num_estados * sizeof(estados));
  if (nfa == NULL) {
    printf("\nERROR, no se ha podido dar memoria a la estructura.");
    exit(0);
  }
  for (j = 0; j < num_estados; j++) {
    fscanf(fich, "%d", &k);			/* lectura estado actual */

    /* memoria para el estado actual */
    nfa[k] = (estados *) malloc(sizeof(estados));
    if (nfa[k] == NULL) {
      printf("\nERROR, no se ha podido dar memoria al registro.");
      exit(0);
    }

    fscanf(fich, "%d", &nfa[k] -> num_transis);	/* lectura del n de transiciones */
    /* memoria para los registros de transiciones, si hay transiciones */
    if (nfa[k] -> num_transis != 0) {
      nfa[k] -> tran = (transicion *) calloc(nfa[k] -> num_transis, sizeof(transicion));
      if (nfa[k] -> tran == NULL) {
	printf("\nERROR, no se ha podido dar memoria a las transiciones.");
	exit(0);
      }

      /* lectura de las transiciones del estado actual */
      for (i = 0; i < nfa[k] -> num_transis; i++) {
	fscanf(fich, " %c", &nfa[k] -> tran[i].entrada);
	fscanf(fich, "%d", &nfa[k] -> tran[i].nuevo_estado);
      }
    }
  }
}

/***************************************************************************/
/* Funcin que realiza la -clausura de un conjunto de estados y devuelve  */
/* en el mismo conjunto su clausura. El conjunto final est formado por    */
/* todos los estados alcanzables con 0 o ms -transiciones.               */
/***************************************************************************/
void Clausura(PILA *p, set *conjunto) {
register unsigned i;
unsigned t;			/* estado a tratar */

  /* introduccin en la pila de todos los estados a clausurar */
  for (i = 0; i < num_estados; i++)
    if (cnjper(conjunto, i))
      push(p, i);

  while (!PILA_VACIA(p -> s)) {
    pop(p, &t);		/* extraccin de un elemento */
    for (i = 0; i < nfa[t] -> num_transis; i++)  /* para cada transicin */
      if (nfa[t] -> tran[i].entrada == '')	 /* cuya entrada sea  */
	/* si no est en la clausura el estado al que se transita, */
	/* se introduce en la clausura y en pila para ser tratado despus */
	if (!cnjper(conjunto, nfa[t] -> tran[i].nuevo_estado)) {
	  cnjins(conjunto, nfa[t] -> tran[i].nuevo_estado);
	  push(p, nfa[t] -> tran[i].nuevo_estado);
	}
  }
}

/***************************************************************************/
/* Funcin que calcula las transiciones de un conjunto de estados mediante */
/* la entrada a. Vamos a ayudarnos de un conjunto auxiliar que contendr   */
/* los nuevos estados a los que se transita.                               */
/***************************************************************************/
void move(set *conjunto, char a) {
register unsigned i, j;
boolean acepta;                 /* TRUE si el estado actaul acepta la entrada */
set *aux;			/* conjunto auxiliar */

  aux = cnjini(num_estados);		/* memoria para el conjunto auxiliar */

/* para cada estado que pertenece al conjunto inicial se calcula su transicin */
  for (i = 0; i < num_estados; i++)
    if (cnjper(conjunto, i)) {
      acepta = FALSE;
      /* ver si el estdo actual acepta la entrada */
      for (j = 0; j < nfa[i] -> num_transis && !acepta; j++)
	if (nfa[i] -> tran[j].entrada == a)
	  acepta = TRUE;
      j--;

      if (acepta) 		/* caracter de entrada vlido */
	cnjins(aux, nfa[i] -> tran[j].nuevo_estado);
    }

  /* se vacia el conjunto inicial y se copia en l el conj. de estados a */
  /* los que se ha transitado (conjunto auxiliar) */
  cnjvac(conjunto);
  cnjcpy(aux, conjunto);
  cnjlib(aux);
}

/***************************************************************************/
/* Funcin que simula el funcionamiento de un NFA. Calcula la transicin   */
/* de un conjunto actual de estados en dos etapas: primero determina todos */
/* los estados alcanzables mediante un entrada a (move(conj, a)) y a       */
/* continuacin calcula la -clausura del nuevo conjunto.                  */
/***************************************************************************/
void Simulacion(void) {
register unsigned i;
char cadena[MAXSTR];		/* cadena de entrada */
unsigned longitud;		/* longitud de la cadena de entrada */
set *conj;
PILA *p;

  p = initPila(num_estados);
  conj = cnjini(num_estados);		/* memoria para el conjunto */
  while (!feof(fich)) {
    fscanf(fich, "%s", cadena);		/* lectura de la cadena */
    cnjins(conj, e_inicial);		/* se inserta el estado inicial */
    Clausura(p, conj);                    /* se calcula la -clausura */
    longitud = strlen(cadena);
    printf("\n\nCadena: %s\n", cadena);
    printf("Estados actuales --> entrada --> Nuevos estados\n");

    /* para cada caracter de entrada de la cadena */
    for (i = 0; i < longitud; i++) {
      Mostrar(*conj);
      /* se calcula la transicin del conjunto de estados para ese caracter */
      move(conj, cadena[i]);
      printf(" --> %c --> ", cadena[i]);
      Mostrar(*conj);
      printf("\n");
      /* se calcula la -clausura */
      Clausura(p, conj);
    }
    printf("\nEstados finales ");
    Mostrar(*conj);
    printf("\nEstado de acptacin: %d", e_aceptacion);

/* si el nico estado de aceptacin est en el conjunto, se acepta la cadena */
    if (cnjper(conj, e_aceptacion))
      printf("\nCADENA ACEPTADA.");
    else
      printf("\nCADENA NO ACEPTADA.");
    Pausa();
    cnjvac(conj);
  }
  Libera_pila(p);
  cnjlib(conj);
}

/***************************************************************************/
int main(int argc, char *argv[]) {

  clrscr();
  if (argc != 3){
    printf("\nERROR, forma de uso: archivo_de_definicin.NFA archivo_de_entrada.IN\n");
    exit(0);
  }
  Abrir(argv[1]);	      /* apertura fichero de definicin */
  Leer_NFA();		      /* caractersticas del autmata */
  Visualizar();
  Abrir(argv[2]);             /* apertura del fichero de entrada de cadenas */
  Simulacion();               /* simulacin del NFA */
  return 0;
}

