/****************************************************************************
 Prctica n 8
	Clculo de los FIRST y FOLLOW para los smbolos de una gramtica

 Fecha: 10 Diciembre 1996

 Autor: Oscar Martnez Valpuesta

 Comentarios:   La gramtica se lee desde un fichero y la estructura que
	      la soporta es la siguiente:
		A cada smbolo de la gramtica se le asocia un nmero, los
	      terminales se numeran comenzando en 0 hasta el nmero de
	      terminales - 1, los no terminales se numeran desde el numero
	      de terminales hasta el nmero de terminales + nmero de no
	      terminales - 1.
		Cada una de las alternativas se representa como un vector
	      de smbolos, acabado en FINALT. Las reglas asociadas a un no
	      terminal se representan como un vector de punteros a dichas
	      alternativas, donde el ltimo elemento es NULL. Existe un
	      vector de punteros a dichas reglas que contiene un elemento
	      por cada no terminal: ese vector representa la gramtica.
   Elemento:								    			Representacin:
   *********								    				==============
   - Smbolo gramatical --> Un entero
   (terminal o no terminal)
   - Alternativa = sucesin de smbolos --> Vector (dinmico) de enteros
   - Gramtica = Conjunto de producciones --> Vector (dinmico) indexado en
					      los no terminales (Habr una
					      serie de alternativas para
					      cada smbolo no terminal).

		A continuacin se van a calcular los no terminales anulables
	      ya que se necesitar para hallar los conjuntos FIRST y FOLLOW
	      saber qu smbolos pueden derivar en . Estos elementos se
	      guardan en una estructura de tipo conjunto.
		El clculo de los FIRST parte de una tabla ya inicializada
	      con los FIRST de los terminales y los de aquellos no terminales
	      que alcanzan  en una derivacin.
/****************************************************************************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "cnjlib.h"
#include "util.h"

#define FINALT  -1    /* Smbolo para indicar el fin de la alternativa en memoria */
#define EPSILON ''   /* Caracter asociado a la cadena vaca en el fichero de entrada */
#define FINF '$'      /* Carcter de final de fichero */
#define MAXSIM  50    /* Tamao del buffer temporal de smbolos */
#define MAXBASURA 100  /* Tamao del vector para leer caracteres superfluos */

#define IN(S, LI, LS) ((S >= LI) && (S < LS))   /* Indica si un smbolo se encuentra en un intervalo dado */

/* Tipos */
typedef int simbolo;          /* Tipo para almacenar un smbolo terminal o no */
typedef simbolo *alternativa; /* Tipo para almacenar una alternativa */
typedef alternativa *regla;   /* Tipo para almacenar varias alternativas */
			      /* asociadas a un mismo no terminal (reglas) */

/* Variables globales */
regla *gram;		     /* Representacin de la gramtica */
unsigned num_term,	     /* Nmero de smbolos terminales */
	 num_no_term,	     /* Nmero de smbolos No terminales */
	 num_simb;	     /* Nmero total de smbolos de la gramtica */
char *alfabeto;		     /* Vector conteniendo todos los smbolos */
char FIRST[15][20];          /* vector de cadenas para los conjuntos FIRST */
			     /*	de los smbolos */
char FOLLOW[15][20];	     /* vector de cadenas para los conjuntos FOLLOW */
			     /*	de los smbolos */
set *anulable;		     /* conjunto de los smbolos anulables */

/* Prototipos */
simbolo sim_to_num(char car);
alternativa *copia_buff(simbolo *buf, unsigned nalt, unsigned nsim);
regla *LeeGram(FILE *fd, unsigned *n_simb, unsigned *num_term,
	       unsigned *num_no_term, char **alfabeto);
void Anulables(void);
void Crear_tabla(void);
void Union(char *a, char *b, boolean *c);
void First(void);
void Follow(void);
void Visualiza(void);
int main(int argc, char *argv[]);

/*****************************************************************************
* sim_to_num --    Devuelve el nmero asociado a un smbolo terminal o no.
*****************************************************************************/
simbolo sim_to_num(char letra) {
int i;
char *ch;

  for (i = 0, ch = alfabeto; (i < num_simb) && (*ch != letra); i++, ch++);
  return ((i < num_simb) ? i : -1);
}


/*****************************************************************************
* copia_buff --     Copia desde un buffer que contiene las alternativas de un
*                   no terminal a la estrucutra de gramatica.
*
* Parmetros:
*		nalt: nmero de alternativas de la produccin.
*		nsim: nmero total de smbolos en la alternativa.
* Comentario:
*		Devuelve NULL en caso de que haya problemas o bien el puntero a la
*		alternativas creadas si todo funciona correctamente.
*****************************************************************************/
alternativa *copia_buff(simbolo *buf, unsigned nalt, unsigned nsim) {
alternativa *altern,	    /* Alternativa de destino */
	    *p_altern;	    /* Puntero para recorrer la alternativa */
simbolo	*cad_simb,          /* Cadena de smbolos de la alternativa */
	*p_cad_simb;	    /* Puntero para recorrer la cadena */

  altern = (alternativa *) calloc(nalt, sizeof(alternativa));
  if (altern == NULL)
    return (NULL);

  cad_simb = (simbolo *)calloc (nsim, sizeof(simbolo));
  if (cad_simb == NULL)
    return (NULL);

  /* Para todas las alternativas */
  for (p_cad_simb = cad_simb, p_altern = altern; nalt--; ) {
    *p_altern++ = p_cad_simb;
    while (*buf != FINALT)
      *p_cad_simb++ = *buf++;		/* Copiar el buffer en la cadena */
    *p_cad_simb++ = *buf++;
  }
  *p_altern = NULL;			/* Marcar el final de la alternativa */

  return (altern);
}


/*****************************************************************************
* LeeGram -- Funcin que lee desde un fichero una gramtica y genera una
*	     estructura en memoria de dicha gramtica.
*
* Parmetros:
*  FILE *fd;		   Fichero de entrada, debe de estar previamente abierto
*  unsigned *n_simb;	   Nmero total de smbolos
*  unsigned *n_term;	   Nmero de smbolos terminales de la gramtica
*  unsigned *n_no_term;	   Nmero de smbolos no terminales
*  char **alfa;		   Vector con los caracteres asociados a cada smbolo
*****************************************************************************/
regla *LeeGram(FILE *fd, unsigned *n_simb, unsigned *n_term, unsigned *n_no_term, char **alfa) {
unsigned i,			/* Contador */
	 n_simb_alt,		/* Nmero de smbolos en la alternativa */
	 num_alt;		/* Nmero total de alternativas */
char  *ch,			/* Carcter ledo del fichero */
      basura[MAXBASURA];	/* Para leer caracteres superfluos */
regla *gramatica;		/* Representacin de la gramtica */
simbolo	*buf,                   /* Buffer temporal para producciones */
	*buf_ptr,		/* Puntero auxiliar al buffer */
	nt;               	/* No terminal en curso de proceso */

/* Lee el nmero de smbolos (terminales + no terminales) */
  fscanf(fd, "%u\n", n_simb);
  *alfa = (char *)calloc(*n_simb, sizeof(char));
  if (*alfa == NULL) {
    printf("\nERROR, no se ha podido dar memoria al alfabeto.");
    exit(0);
  }
  ch = *alfa;		/* Para dejar los smbolos en el vector alfa */

/* Lee los terminales */
  fscanf(fd, "%[ \t]", basura);
  for (i = 0; (i < *n_simb) && (fscanf(fd, "%[\n]", basura) != 1); i++, ch++)
    fscanf(fd, "%c%[ \t]", ch, basura);
  *n_term = i;

/* Lee los no terminales */
  fscanf(fd, "%[ \t]", basura);
  for (; (fscanf(fd, "%[\n]", basura) != 1) && (i < *n_simb); i++, ch++)
    fscanf(fd, "%c%[ \t]", ch, basura);
  if (i != *n_simb) {
    printf("\nERROR, nmero de smbolos leidos no coincide con los esperados.");
    exit(0);
  }
  *n_no_term = *n_simb - *n_term;

/* Lee las reglas */
  gramatica = (regla *)calloc(*n_no_term, sizeof(regla));
  buf = (simbolo *)calloc(MAXSIM, sizeof(simbolo));
  if ((buf == NULL) || (gramatica == NULL)) {
    printf("\nERROR, no se ha podido dar memoria.");
    exit(0);
  }
  for (i = 0; (i < *n_no_term); i++) {
  /* Lee parte izquierda de las producciones */
    fscanf(fd, "%[ \t]%c", basura, ch);
    nt = sim_to_num(*ch);
    if (!IN(nt, (int)*n_term, (int)*n_simb)) {
      printf("\nERROR, ledo un smbolo distinto de los no terminales en la izquierda de una regla");
      exit(0);
    }
    fscanf(fd, "%[ \t]->", basura);

    /* Lee parte derecha de las producciones */
    buf_ptr = buf;
    num_alt = 0;
    n_simb_alt = 0;
    do {
      fscanf(fd, "%[ \t]", basura);
      do {
	fscanf(fd, "%c%[ \t]", ch, basura);
	if (*ch != EPSILON) {
	  *buf_ptr = sim_to_num(*ch);
	    if (!IN(*buf_ptr, (int)0, (int)*n_simb)) {
	      printf("\nERROR, Ledo un smbolo desconocido");
	      exit(0);
	    }
	    buf_ptr++;
	    n_simb_alt++;
	}
      } while (fscanf(fd, "%[|\n]", basura) == 0);
      *buf_ptr++ = FINALT;		/* Indicando fin de alternativa */
      n_simb_alt++;
      num_alt++;
    } while (*basura == '|');
    if ((gramatica[(unsigned)nt - *n_term] = copia_buff(buf, num_alt, n_simb_alt)) == NULL) {
      printf("\nERROR, no se ha podido dar memoria1.");
      exit(0);
    }
  }
  free(buf);					/* El buffer era slo local */
  return (gramatica);
}

/****************************************************************************/
/* Funcin que devuelve el conjunto de smbolos que son capaces de llegar a */
/*  en finitas derivaciones. Al principio, ese conjunto est formado por   */
/* aquellos no terminales que alcanzan  en una derivacin.
/****************************************************************************/
void Anulables(void) {
register unsigned i, j, k;
unsigned num_anulables = 0;
boolean change, salir;

  anulable = cnjini(num_simb);
  /* marcar todo los no terminales con una alternativa  */
  for (i = 0; i < num_no_term; i++)
    for (j = 0; gram[i][j] != NULL && !cnjper(anulable, num_term + i); j++)
      if (gram[i][j][0] == FINALT) {
	cnjins(anulable, num_term + i);
	num_anulables++;
      }
/* si hay tantos elementos anulables como no terminales no hace falta seguir */
  if (num_anulables != num_no_term)
    do {
      change = FALSE;
      for (i = 0; i < num_no_term; i++)
	if (!cnjper(anulable, num_term + i))
	  for (j = 0; gram[i][j] != NULL; j++) {
	    salir = FALSE;
	    for (k = 0; gram[i][j][k] != FINALT && !salir; k++)
/* si algn elemento no deriva en , no miramos el resto de la regla */
	      if (!cnjper(anulable, gram[i][j][k]))
		salir = TRUE;
/* si hemos llegado al final de la regla y todo derivan en , se almacena */
	    if (!salir) {
	      cnjins(anulable, num_term + i);
	      change = TRUE;
	    }
	  }
    } while (change);
}

/****************************************************************************/
/* Funcin que el  FIRST inicial de los smbolos y el FOLLOw de los         */
/* terminales. El de los terminales es l mismo (FIRST(a) = 'a'y el de los  */
/* no terminales va a ser  si la regla puede derivar en .                 */
/****************************************************************************/
void Crear_tabla(void) {
register unsigned i;


  /* el FIRST y el FOLLOW de un no terminal es l mismo */
  for (i = 0; i < num_term; i++) {
    *FIRST[i] = alfabeto[i];
    *FOLLOW[i] = alfabeto[i];
  }
  /* las reglas que pueden derivar en  (anulable), lo contienen */
  for (i = num_term; i < num_simb; i++)
    if (cnjper(anulable, i))
      *FIRST[i] = EPSILON;
}

/****************************************************************************/
/* Procedimiento que calcula la unin de dos conjuntos de caracteres. El    */
/* se devuelve en el primer parmetro y la variable c indica si hubo algn  */
/* cambio.                                                                  */
/****************************************************************************/
void Union(char *a, char *b, boolean *c) {
unsigned length1, length2;     /* longitudes de los conjuntos */
register unsigned i, j;
boolean encontrado;    /* indica si el carcter actual est en el otro conj */
char aux[1];

  length1 = strlen(a);
  length2 = strlen(b);
/* se compara cada elemento de la segunda cadena con todos los de la primera */
  for (i = 0; i < length2; i++) {
    encontrado = FALSE;
    for (j = 0; j < length1 && !encontrado; j++)
      if (b[i] == a[j])
	encontrado = TRUE;

    if (!encontrado)
      if (b[i] != EPSILON) {
	aux[0] = b[i];
	aux[1] = '\0';
	strcat(a, aux);
	*c = TRUE;
      }
  }
}

/****************************************************************************/
/* Algoritmo que calcula el FIRST de todos los smbolos de una gramtica.   */
/* Primero calcula el FIRST para cada terminal y para cada no terminale que */
/* en una derivacin alcance a .                                           */
/****************************************************************************/
void First(void) {
boolean change, cambio;
register unsigned i, j, k;

  Anulables();		/* calcula los anulables */
  Crear_tabla();        /* iniciamos los FIRST y FOLLOW */
  do {
    change = FALSE;
    for (i = 0; i < num_no_term; i++)
      for (j = 0; gram[i][j] != NULL; j++) {
	cambio = FALSE;
	for (k = 0; gram[i][j][k] != FINALT && !cambio; k++)
	  /* si no hay recursividad */
	  if (alfabeto[gram[i][j][k]] != alfabeto[i+num_term]) {
	    Union(FIRST[i + num_term], FIRST[gram[i][j][k]], &change);
	    /* si no es anulable, acabamos con esta regla */
	    if (!cnjper(anulable, gram[i][j][k]))
	      cambio = TRUE;
	  }
	  else
	    if (!cnjper(anulable, gram[i][j][k]))
	      cambio = TRUE;
	/* si hay recursividad y es anulable sigue iterando */
      }
  } while (change);
}

/****************************************************************************/
/* Funcin que muestra en pantalla la gramtica.                            */
/****************************************************************************/
void Visualiza(void) {
register unsigned i,j,k;

  for (i = 0; i < num_no_term; i++) {
    for (j = 0; gram[i][j] != NULL; j++) {
      printf("\n%c -->", alfabeto[num_term+i]);
      for (k = 0; gram[i][j][k] != FINALT; k++)
	printf("[%c]", alfabeto[gram[i][j][k]]);
      printf("[%d]", gram[i][j][k]);
    }
  }
}

/*****************************************************************************
 Funcin que calcula los FOLLOW de todo los no terminales. Se basa en el
 algoritmo descrito en libro de Aho, Sethi y Ullman "Compiladores. Principios,
 tcnicas y herramientas." (Addison Wesley, 1986):
 1. Pngase $ (FINF) en FOLLOW(S), donde S es el smbolo de arranque.
 2. Si hay una produccin A -> B, todo lo que est en FIRST() excepto  se
    pone en FOLLOW(B).
 3. Si hay una produccin A -> B o una produccin A -> B, donde FIRST()
    contenga a , entonces todo lo que est en FOLLOW(A) se pone en FOLLOW(B).
*****************************************************************************/
void Follow(void) {
register unsigned i, j, k;
boolean change;

  *FOLLOW[num_term] = FINF;  /* 1. */
  do{
    change = FALSE;
    for (i = 0; i < num_no_term; i++)
      for (j = 0; gram[i][j] != NULL; j++)
	for (k = 0; gram[i][j][k] != FINALT;  k++) {
	  if (gram[i][j][k+1] != FINALT) /* 2. */
	    Union(FOLLOW[gram[i][j][k]], FIRST[gram[i][j][k+1]], &change);

/* 3. */  if (cnjper(anulable, gram[i][j][k+1]) || gram[i][j][k+1] == FINALT)
	    Union(FOLLOW[gram[i][j][k]], FOLLOW[i + num_term], &change);

	  if (cnjper(anulable, gram[i][j][k]) && gram[i][j][k+1] != FINALT)
	    Union(FOLLOW[gram[i][j][k]], FOLLOW[gram[i][j][k+1]], &change);
	}
  } while (change);

}

/****************************************************************************/
int main(int argc, char *argv[]) {
FILE * fd;
register unsigned i;

  if (argc != 2) {
    printf("Modo de uso: pract8 <fichero.GRA>\n");
    exit(0);
  }
  fd = fopen(argv[1], "r");
  if (fd == NULL) {
    printf("\nERROR, no se ha podido abrir el fichero %s.", argv[1]);
    exit(0);
  }

  gram = LeeGram(fd, &num_simb, &num_term, &num_no_term, &alfabeto);

  First();
  Follow();

  for (i = num_term; i < num_simb; i++)
    printf("\nFIRST de %c = {%s}   FOLLOW de %c = {%s}",alfabeto[i], FIRST[i], alfabeto[i], FOLLOW[i]);
  Pausa();

  return(0);
}


