/****************************************************************************
 Prctica n 9	   Analizador sintctico descendente recursivo para
		   expresiones regulares.

 Fecha: 8 Enero 1997

 Autor: Oscar Martnez Valpuesta

 Objetivo:   Se pretende aplicar el mtodo de anlisis sintctico descendente
	   recursivo para reconocer expresiones regulares.
 Comentarios: La gramtica de las expresiones regulares que se utilizar es:

	    Expresin --->Trmino Resto_expresin
	    Resto_expresin ---> | Expresin
	    Resto_expresin ---> 
	    Trmino ---> Cotrmino Resto_trmino
	    Resto_trmino ---> Trmino
	    Resto_trmino --> 
	    Cotrmino ---> Factor Resto_cotrmino
	    Resto_cotrmino ---> * Resto_cotrmino
	    Resto_cotrmino ---> 
	    Factor ---> (Expresin)
	    Factor ---> id
	    Factor ---> &

	    El smbolo & indica la expresin regular vaca y los identificadores
	    son de una letra.

	      Se ha diseado una funcin para cada smbolo no terminal de
	    la gramtica que funcionar como una regla de produccin. Tambin
	    existe una funcin que almacena los first de los no terminales
	    ya que estos pueden ser usados en algn momento. Como la gramtica
	    es conocida estos son introducidos de manera manual y no hace
	    falta aplicar el algortmo visto en clase.

	      Las cadenas de entrada se leen desde un fichero de texto pasado
	    por parmetro al programa.
/****************************************************************************/
#include <string.h>    /* strlen */
#include <stdlib.h>
#include "util.h"

#define MAX 40	   /* n de elementos como mximo en la cadena de entrada */
#define NO_TERM 7  /* nmero de no terminales */
#define ALFABETO "abcdefghijklmnopqrstuvwxyz"

/******************************* TIPOS *************************************/


/****************************** VARIABLES **********************************/
FILE *fichero;		/* handled para el fichero de entrada */
char cadena[MAX];       /* cadena de entrada */
unsigned posicion;	/* posicin actual en la cadena de entrada */
unsigned longitud;	/* longitud de la cadena */
boolean error;		/* indica si se ha producido un error */
char lookahead;		/* smbolo de preanlisis */
char first[NO_TERM][MAX];


/***************************** PROTOTIPOS **********************************/
void Abrir(FILE **f, char s[]);
void Leer_cadena(char *s);
char get_token(unsigned p);
void match(char m);
boolean In(char string[], char c);
void Calcula_F(void);
void Expresion(void);
void Termino(void);
void Resto_expresion(void);
void Cotermino(void);
void Resto_termino(void);
void Factor(void);
void Resto_cotermino(void);

/***************************************************************************/
/* Funcin que abre un fichero de alto nivel para lectura.                 */
/***************************************************************************/
void Abrir(FILE **f, char s[]) {

  *f = fopen(s, "r");
  if (*f == NULL) {
    ERROR1(s);
    exit(0);
  }
}

/***************************************************************************/
/*  Funcin que lee desde un fichero de texto una cadena de entrada y      */
/* almacena en una variable global la longitud de la misma.                */
/***************************************************************************/
void Leer_cadena(char *s) {

  fscanf(fichero, "%s", s);
  longitud = strlen(s);
  posicion = 0;
}

/***************************************************************************/
/* Funcin que devuelve el token de la posicin p en la cadena de entrada. */
/***************************************************************************/
char get_token(unsigned p) {
char token;

  token = cadena[p];
  return token;
}

/***************************************************************************/
/*  Funcin que actualiza el smbolo de preanlisis siempre y cuando no se */
/* haya producido un error y coincida el carcter esperado con el lookahead*/
/* actual.         							   */
/***************************************************************************/
void match(char m) {

  if (lookahead == m && !error)
    lookahead = get_token(++posicion);
  else
    error = TRUE;
}

/***************************************************************************/
/*  Funcin que devuelve si un carcter est contenido en un string.       */
/* Parmetros:                                                             */
/* c: carcter a buscar.                                                   */
/* string: string en el que se busca el carcter c.                        */
/***************************************************************************/
boolean In(char string[], char c) {
unsigned length;
boolean encontrado;
register unsigned i;

  encontrado = FALSE;
  length = strlen(string);
  for (i = 0; i < length && !encontrado; i++)
    if (c == string[i])
      encontrado = TRUE;

  return encontrado;
}

/***************************************************************************/
/*  Funcin que almacena los first de los smbolos no terminales.          */
/***************************************************************************/
void Calcula_F(void) {

  strcpy(first[0], "(&");		/* Expresion */
  strcat(first[0], ALFABETO);
  strcpy(first[1], first[0]);           /* Termino */
  strcpy(first[2], "|");		/* Resto_expresion */
  strcpy(first[3], first[0]);		/* Cotermino */
  strcpy(first[4], "");
  strcat(first[4], first[0]);		/* Resto_termino */
  strcpy(first[5], first[0]);		/* Factor */
  strcpy(first[6], "*");		/* Resto_cotermino */

}

/***************************************************************************/
void Expresion(void) {

  if (!error)
    Termino();
  if (!error)
    Resto_expresion();
}

/***************************************************************************/
void Termino(void) {

  if (!error)
    Cotermino();
  if (!error)
    Resto_termino();
}

/***************************************************************************/
void Resto_expresion(void) {

  if (lookahead == '|' && !error) {
    match('|');
    Expresion();
  }
  else ;	/* palabra vaca */
}

/***************************************************************************/
void Cotermino(void) {

  if (!error)
    Factor();
  if (!error)
    Resto_cotermino();
}

/***************************************************************************/
void Resto_termino(void) {

  if (In(first[1],lookahead) && !error)
    Termino();
  else
    if (!error);
}

/***************************************************************************/
void Factor(void) {

  if (lookahead == '(' && !error) {
    match('(');
    Expresion();
    match(')');
  }
  else
    if (lookahead == '&' && !error)
      match('&');
    else
      if (In(ALFABETO, lookahead))
	match(lookahead);
      else
	if (!error)
	  error= TRUE;
}

/***************************************************************************/
void Resto_cotermino(void) {

  if (lookahead == '*' && !error)
    if (cadena[posicion-1] != '*') {
      match('*');
      Resto_cotermino();
    }
    else
      error = TRUE;
  else;
}

/***************************************************************************/
int main(int argc, char *argv[]) {

  clrscr();
  if (argc != 2){
    printf("\nERROR, forma de uso: ejecutable <archivo_de_entrada.IN>\n");
    exit(0);
  }
  Abrir(&fichero, argv[1]);	/* apertura del fichero de datos */
  Calcula_F();				/* clculo de los first */
  do {
    Leer_cadena(cadena);	/* leemos una cadena de entrada */
    lookahead = get_token(posicion);  /* inicializacin del lookahead */
    error = FALSE;
    Expresion();		      /* smbolo de arranque */
    if (!error && posicion == longitud) {
      printf("\n\n%s\tExpresin regular correcta.", cadena);
      Pausa();
    }
    else {
      printf("\n\n%s\tExpresin regular incorrecta, error en la posicin %d.", cadena, posicion+1);
      Pausa();
    }
  } while (!feof(fichero));

  fclose(fichero);
  return 0;
}
